"""
INN trainer
----------------------------------------------------------------------------------
Copyright (C) 2024  Chanwook Park
 Northwestern University, Evanston, Illinois, US, chanwookpark2024@u.northwestern.edu
"""
from pyinn import dataset_classification, dataset_regression, model, train, plot # with pyinn library
# import dataset_classification, dataset_regression, model, train, plot # for debugging
from jax import config
config.update("jax_enable_x64", True)
import os
import yaml

# %% User Set up
with open('./pyinn/settings.yaml','r') as file:
    settings = yaml.safe_load(file)

gpu_idx = settings['GPU']['gpu_idx']  # set which GPU to run on Athena
os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"  # GPU indexing
os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_idx)  # GPU indexing

run_type = settings['PROBLEM']["run_type"]
interp_method = settings['PROBLEM']["interp_method"]
data_name = settings['DATA']["data_name"]

with open(f'./config/{data_name}.yaml','r') as file_dataConfig:
    config = yaml.safe_load(file_dataConfig)
    config['interp_method'] = settings['PROBLEM']["interp_method"]
    config['TD_type'] = settings['PROBLEM']["TD_type"]
    

# --------------------- Regression --------------------------3
if run_type == "regression":
    
    ## data import
    data = dataset_regression.Data_regression(data_name, config)

    ## train
    if interp_method == "linear" or interp_method == "nonlinear":
        regressor = train.Regression_INN(data, config)
    elif interp_method == "MLP":
        regressor = train.Regression_MLP(data, config)
    regressor.train()

    ## plot
    plot.plot_regression(regressor, data, config)

# --------------------- Classification --------------------------
elif run_type == "classification": 

    ## data import
    data = dataset_classification.Data_classification(data_name, config)
    
    ## train
    if interp_method == "linear" or interp_method == "nonlinear":
        classifier = train.Classification_INN(data, config)
    elif interp_method == "MLP":
        classifier = train.Classification_MLP(data, config)  
    classifier.train()

    ## plot
    plot.plot_classification(classifier, data, config)




    





"""
INN trainer
----------------------------------------------------------------------------------
Copyright (C) 2024  Chanwook Park
 Northwestern University, Evanston, Illinois, US, chanwookpark2024@u.northwestern.edu
"""

import jax
import jax.numpy as jnp
from jax.nn.initializers import uniform
from jax import config
config.update("jax_enable_x64", True)
from functools import partial
from typing import (Any, Callable, Iterable, List, Optional, Sequence, Tuple,
                    Union)
from jax.scipy.interpolate import RegularGridInterpolator
from .Interpolator import LinearInterpolator, NonlinearInterpolator ## when using pyinn
# from Interpolator import LinearInterpolator, NonlinearInterpolator ## when debugging

class INN_linear:
    def __init__(self, grid_dms, config):
        """ 1D linear interpolation
        --- input --- 
        grid_dms: (dim,J) 1D vector of the grid and dimension
        values: (J,) 1D vector of nodal values
        """
        self.grid_dms = grid_dms
        self.config = config
        # self.interpolate = LinearInterpolator(grid)
        
    @partial(jax.jit, static_argnames=['self'])
    def get_Ju_idata_imd_idm_ivar(self, x_idata_idm, grid, u_imd_idm_ivar_nds):
        """ compute interpolation for a single mode, 1D function
        --- input ---
        x_idata_idm: scalar, jnp value / this can be any input
        grid: (J,) 1D vector of the grid
        u_imd_idm_ivar_nds: (J,) jnp 1D array
        --- output ---
        Ju_idata_imd_idm_ivar: scalar, 1D interpolated value
        """
        interpolate = LinearInterpolator(grid)
        # Ju_idata_imd_idm_ivar = self.interpolate(x_idata_idm, u_imd_idm_ivar_nds)
        Ju_idata_imd_idm_ivar = interpolate(x_idata_idm, u_imd_idm_ivar_nds)
        return Ju_idata_imd_idm_ivar
    get_Ju_idata_imd_idm_vars = jax.vmap(get_Ju_idata_imd_idm_ivar, in_axes = (None,None,None,0)) # input: scalar, (J,), (var,J) / output: (var,)
    get_Ju_idata_imd_dms_vars = jax.vmap(get_Ju_idata_imd_idm_vars, in_axes = (None,0,0,0)) # input: (dim,), (dim,J) (dim,var,J) / output: (dim,var)
    get_Ju_idata_mds_dms_vars = jax.vmap(get_Ju_idata_imd_dms_vars, in_axes = (None,None,None,0)) # input: (dim,), (dim,J), (M,dim,var,J) / output: (M,dim,var)

    ## CP decomposition
    # def get_Ju_idata_imd(self, x_idata_dms, u_imd_dms_vars_nds):
    #     # x_idata_dms: (dim,)
    #     # x_dms_nds: (dim, nnode)
    #     # u_imd_dms_vars_nds: (dim, var, nnode)
        
    #     Ju_idata_imd_dims_vars = self.get_Ju_idata_imd_dms_vars(x_idata_dms, u_imd_dms_vars_nds) # output: (dim, var)
    #     Ju_idata_imd = jnp.prod(Ju_idata_imd_dims_vars, axis=0)
    #     return Ju_idata_imd # (var,)
    # get_Ju_idata_mds = jax.vmap(get_Ju_idata_imd, in_axes = (None,None,0)) # output: (mds,var)

    # def get_Ju_idata(self, x_idata_dms, u_mds_dms_vars_nds):
    #     Ju_idata_mds = self.get_Ju_idata_mds(x_idata_dms, u_mds_dms_vars_nds) # (mds,var)
    #     Ju_idata = jnp.sum(Ju_idata_mds, axis=0) # returns (var,)
    #     return Ju_idata

    # # @partial(jax.jit, static_argnames=[]) # jit necessary
    # # @jax.jit
    # @partial(jax.jit, static_argnames=['self'])
    # def forward(self, params, x_idata):
    #     """ Prediction function
    #         run one forward pass on given input data
    #         --- input ---
    #         params: u_mds_dms_vars_nds, (nmode, dim, var, nnode)
    #         x_dms_nds: nodal coordinates (dim, nnode)
    #         x_idata: x_idata_dms (dim,)
    #         --- return ---
    #         predicted output (var,)
    #     """
    #     pred = self.get_Ju_idata(x_idata, params)
    #     return pred
    # v_forward = jax.vmap(forward, in_axes=(None,None, 0)) # returns (ndata,)
    # vv_forward = jax.vmap(v_forward, in_axes=(None,None, 0)) # returns (ndata,)

    def tucker(self, G, factors):
        """ serior computation of tucker decomposition 
        --- input ---
        G: core tensor, (M, M, ..., M) 
        factors: factor matrices, (dim, M)"""
        for factor in factors:
            G = jnp.tensordot(G, factor, axes=[0,0])
        return jnp.squeeze(G)
    v_tucker = jax.vmap(tucker, in_axes=(None,None,0)) # returns (var,)

    @partial(jax.jit, static_argnames=['self'])
    def forward(self, params, x_idata):
        """ Prediction function
            run one forward pass on given input data
            --- input ---
            params: u_mds_dms_vars_nds, (nmode, dim, var, nnode)
            x_dms_nds: nodal coordinates (dim, nnode)
            x_idata: x_idata_dms (dim,)
            --- return ---
            predicted output (var,)
        """
        # if self.config['TD_type']=='CP':
        pred = self.get_Ju_idata_mds_dms_vars(x_idata, self.grid_dms, params) # input: (dim,), (dim,J), (M,dim,var,J) / output: (M,dim,var)
        pred = jnp.prod(pred, axis=1) # output: (M,var)
        pred = jnp.sum(pred, axis=0) # output: (var,)
        
        # if self.config['TD_type']=='Tucker':
            
        #     G = params[0]
        #     factors = self.get_Ju_idata_mds_dms_vars(x_idata, params[1]) # output: (M,dim,var)
        #     # for factor in pred.transpose(1,0,2): # factor: (M,var)
        #     #     G = jnp.tensordot(G, factor, axes=[0,0])
        #     factors = factors.transpose(2,1,0) # (var,dim,M)
        #     # print(factors.shape)
        #     pred = self.v_tucker(G, factors)

        return pred 

    v_forward = jax.vmap(forward, in_axes=(None,None, 0)) # returns (ndata,)
    vv_forward = jax.vmap(v_forward, in_axes=(None,None, 0)) # returns (ndata,)
        

class INN_nonlinear(INN_linear):
    def __init__(self, grid_dms, config):
        super().__init__(grid_dms, config) # prob being dropout probability

        self.nelem = config['MODEL_PARAM']['nelem']
        self.nnode = self.nelem + 1
        self.s_patch = config['MODEL_PARAM']['s_patch']
        self.alpha_dil = config['MODEL_PARAM']['alpha_dil'] 
        self.p_order = config['MODEL_PARAM']['p_order']
        p_dict = {0:0, 1:2, 2:3, 3:4, 4:5, 5:6} 
        self.mbasis = p_dict[self.p_order] 
        self.radial_basis = config['MODEL_PARAM']['radial_basis']
        self.activation = config['MODEL_PARAM']['INNactivation']

        # self.interpolate = NonlinearInterpolator(grid, self.config)

    @partial(jax.jit, static_argnames=['self']) #    , 's_patch', 'alpha_dil', 'p_order', 'radial_basis', 'INNactivation'])
    def get_Ju_idata_imd_idm_ivar(self, x_idata_idm, grid, u_imd_idm_ivar_nds):
        """ compute interpolation for a single mode, 1D function
        --- input ---
        x_idata_idm: scalar, jnp value / this can be any input
        grid: (J,) 1D vector of the grid
        u_imd_idm_ivar_nds: (J,) jnp 1D array
        --- output ---
        Ju_idata_imd_idm_ivar: scalar, 1D interpolated value
        """
        interpolate = NonlinearInterpolator(grid, 
                                            self.nelem, self.nnode, self.s_patch, self.alpha_dil, self.p_order, 
                                            self.mbasis, self.radial_basis, self.activation)
        # Ju_idata_imd_idm_ivar = self.interpolate(x_idata_idm, u_imd_idm_ivar_nds)
        Ju_idata_imd_idm_ivar = interpolate(x_idata_idm, u_imd_idm_ivar_nds)
        return Ju_idata_imd_idm_ivar
    get_Ju_idata_imd_idm_vars = jax.vmap(get_Ju_idata_imd_idm_ivar, in_axes = (None,None,None,0)) # input: scalar, (J,), (var,J) / output: (var,)
    get_Ju_idata_imd_dms_vars = jax.vmap(get_Ju_idata_imd_idm_vars, in_axes = (None,0,0,0)) # input: (dim,), (dim,J) (dim,var,J) / output: (dim,var)
    get_Ju_idata_mds_dms_vars = jax.vmap(get_Ju_idata_imd_dms_vars, in_axes = (None,None,None,0)) # input: (dim,), (dim,J), (M,dim,var,J) / output: (M,dim,var)

## MLP
def relu(x):
    return jnp.maximum(0, x)

@partial(jax.jit, static_argnames=['activation'])
def forward_MLP(params, activation, x_idata):
    # per-example predictions
    activations = x_idata
    for w, b in params[:-1]:
        outputs = jnp.dot(w, activations) + b
        if activation == 'relu':
            activations = relu(outputs)
        elif activation == 'sigmoid':
            activations = jax.nn.sigmoid(outputs)
    final_w, final_b = params[-1]
    return jnp.dot(final_w, activations) + final_b
v_forward_MLP = jax.vmap(forward_MLP, in_axes=(None,None, 0)) # returns (ndata,)
vv_forward_MLP = jax.vmap(v_forward_MLP, in_axes=(None,None, 0)) # returns (ndata,)


import numpy as np
import jax.numpy as jnp
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import os
from .model import * ## when using pyinn
from .dataset_regression import *
from .dataset_classification import *
# from model import * ## when debugging
# from dataset_regression import *
# from dataset_classification import *


def plot_regression(model, cls_data, config):
    bool_plot = config['PLOT']['bool_plot']
    
    if bool_plot:

        plot_in_axis = config['PLOT']['plot_in_axis']
        plot_out_axis = config['PLOT']['plot_out_axis']


        # make a directory if there isn't any
        if not os.path.exists('plots'):
            os.makedirs('plots')
        
        if len(plot_in_axis)==2 and  cls_data.bool_normalize == False:
            # we will plot the error only when there is no normalization on the original data.
            plot_2D_1D(model, cls_data, plot_in_axis, plot_out_axis)

            if config['interp_method'] != "MLP":
                plot_modes(model, cls_data, plot_in_axis, plot_out_axis)
        
        elif len(plot_in_axis)==1 and len(plot_out_axis)==1 and  cls_data.bool_normalize == False:
            # we will plot the error only when there is no normalization on the original data.
            plot_1D_1D(model, cls_data, plot_in_axis, plot_out_axis)

        elif len(plot_in_axis)==1 and len(plot_out_axis)==2 and  cls_data.bool_normalize == False:
            # we will plot the error only when there is no normalization on the original data.
            plot_1D_2D(model, cls_data, plot_in_axis, plot_out_axis)

        elif len(plot_in_axis)==2 and len(plot_out_axis)==1 and  cls_data.bool_normalize == False:
            # for spiral classification
            plot_2D_classification(model, cls_data, plot_in_axis, plot_out_axis)

        elif len(plot_in_axis)==3 and len(plot_out_axis)==1 and  cls_data.bool_normalize == True:
            # we will plot the error only when there is no normalization on the original data.
            # plot_2D_1D(model, cls_data, [0,1], plot_out_axis)
            if config['interp_method'] != "MLP":
                plot_modes(model, cls_data, plot_in_axis, plot_out_axis)

    else:
        print("\nPlotting deactivated\n")
        pass

def plot_classification(model, cls_data, config):
    bool_plot = config['PLOT']['bool_plot']
    
    if bool_plot:
        plot_in_axis = config['PLOT']['plot_in_axis']
        plot_out_axis = config['PLOT']['plot_out_axis']


        # make a directory if there isn't any
        if not os.path.exists('plots'):
            os.makedirs('plots')
        
        if len(plot_in_axis)==2 and len(plot_out_axis)==1:
            # for spiral classification
            plot_2D_classification(model, cls_data, plot_in_axis, plot_out_axis)
    else:
        print("\nPlotting deactivated\n")
        pass


def plot_1D_1D(model, cls_data, plot_in_axis, plot_out_axis):
    """ This function plots 2D input and 1D output data. By default, this function should work on original data only
    plot_in_axis: [axis1]
    plot_out_axis: [axis1]
    """

    ## create mesh and data
    ### in normalized space, create prediction
    xmin, xmax = cls_data.x_data_minmax["min"][plot_in_axis[0]], cls_data.x_data_minmax["max"][plot_in_axis[0]]
    
    x_nds = jnp.linspace(xmin, xmax, 101, dtype=jnp.float64).reshape(-1,1) # (101,1)
    if model.interp_method == "linear" or model.interp_method == "nonlinear":
        U_pred = model.v_forward(model.params, x_nds) # (101,L)
    elif model.interp_method == "MLP":
        U_pred = model.v_forward(model.params, model.activation, x_nds) # (101,L)
    
    U_exact = globals()["v_fun_"+cls_data.data_name](x_nds) # (101,L)    

    fig = plt.figure(figsize=(6,5))
    gs = gridspec.GridSpec(1, 1)
    ax1 = fig.add_subplot(gs[0])
    # plt.subplots_adjust(wspace=0.4)  # Increase the width space between subplots

    ax1.plot(x_nds, U_exact, '-', color='k', linewidth = 4,  label='Original function')
    ax1.plot(x_nds, U_pred, '-', color='g', linewidth = 4,  label='Prediction')
    ax1.set_xlabel(fr"$x_{str(plot_in_axis[0]+1)}$", fontsize=16)
    ax1.set_ylabel(fr"$u_{str(plot_out_axis[0]+1)}$", fontsize=16)
    ax1.tick_params(axis='both', labelsize=12)
    # ax1.set_title('INN prediction', fontsize=16)
    ax1.legend(shadow=True, borderpad=1, fontsize=14, loc='best')
    plt.tight_layout()

    parent_dir = os.path.abspath(os.getcwd())
    path_figure = os.path.join(parent_dir, 'plots')
    fig.savefig(os.path.join(path_figure, cls_data.data_name + "_" + model.interp_method) , dpi=300)
    plt.close()

def plot_modes(model, cls_data, plot_in_axis, plot_out_axis):
    """ This function plots multiple subplots in a dimension of (M, I). Each subplot represents 1D plot of input i at mode m.
    plot_in_axis: [axis1, axis2, ...]
    plot_out_axis: [axis1]
    """
    ## create mesh and data
    ### in normalized space, create prediction
    TD_type = model.config["TD_type"]

    if TD_type == "CP":
        params = model.params
    elif TD_type == "Tucker":
        params = model.params[1]
        print("Core tensor of Tucker product")
        print(model.params[0]) # print out 

    M = params.shape[0] # nmode
    I = params.shape[1] # dim
    J = params.shape[3] # nnode

    fig, axes = plt.subplots(M, I) 
    for idx, ax in enumerate(axes.flat):
        m = idx // I # m-th mode
        i = idx % I # i-th dimension

        xmin, xmax = cls_data.x_data_minmax["min"][plot_in_axis[i]], cls_data.x_data_minmax["max"][plot_in_axis[i]]
        x_nds = jnp.linspace(xmin, xmax, J, dtype=jnp.float64) # (J,)
        u_nds = params[m,i,plot_out_axis[0],:] # (J,)
        ax.plot(x_nds, u_nds, 'o')
        ax.set_title(f"{m+1}-th mode, {i+1}-th dim")

    plt.tight_layout()

    parent_dir = os.path.abspath(os.getcwd())
    path_figure = os.path.join(parent_dir, 'plots')
    TD_type = model.config['TD_type'] 
    fig.savefig(os.path.join(path_figure, cls_data.data_name + f"_{TD_type}_" + model.interp_method + f"_{M}modes") , dpi=300)
    plt.close()

def plot_1D_2D(model, cls_data, plot_in_axis, plot_out_axis, color_map="viridis", vmin=0, vmax=1, marker_size=20):
    """ This function plots 2D input and 1D output data. By default, this function should work on original data only
    plot_in_axis: [axis1, axis2]
    plot_out_axis: [axis1]
    """
    ## create mesh and data
    ### in normalized space, create prediction
    xmin, xmax = cls_data.x_data_minmax["min"][plot_in_axis[0]], cls_data.x_data_minmax["max"][plot_in_axis[0]]
    
    x_nds = jnp.linspace(xmin, xmax, 101, dtype=jnp.float64).reshape(-1,1) # (101,1)
    if model.interp_method == "linear" or model.interp_method == "nonlinear":
        U_pred = model.v_forward(model.params, x_nds) # (101,L)
    elif model.interp_method == "MLP":
        U_pred = model.v_forward(model.params, model.activation, x_nds) # (101,L)
    
    U_exact = globals()["v_fun_"+cls_data.data_name](x_nds) # (101,L)    

    fig = plt.figure(figsize=(10,5))
    gs = gridspec.GridSpec(1, 2)
    ax1 = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1])
    plt.subplots_adjust(wspace=0.4)  # Increase the width space between subplots

    ax1.plot(x_nds, U_exact[:,[plot_out_axis[0]]], '-', color='k', linewidth = 4,  label='Original function')
    ax1.plot(x_nds, U_pred[:,[plot_out_axis[0]]], '-', color='g', linewidth = 4,  label='Prediction')
    ax1.set_xlabel(fr"$x_{str(plot_in_axis[0]+1)}$", fontsize=16)
    ax1.set_ylabel(fr"$u_{str(plot_out_axis[0]+1)}$", fontsize=16)
    ax1.tick_params(axis='both', labelsize=12)
    # ax1.set_title('INN prediction', fontsize=16)
    ax1.legend(shadow=True, borderpad=1, fontsize=14, loc='best')
    
    ax2.plot(x_nds, U_exact[:,[plot_out_axis[1]]], '-', color='k', linewidth = 4,  label='Original function')
    ax2.plot(x_nds, U_pred[:,[plot_out_axis[1]]], '-', color='g', linewidth = 4,  label='Prediction')
    ax2.set_xlabel(fr"$x_{str(plot_in_axis[0]+1)}$", fontsize=16)
    ax2.set_ylabel(fr"$u_{str(plot_out_axis[1]+1)}$", fontsize=16)
    ax2.tick_params(axis='both', labelsize=12)
    # ax1.set_title('INN prediction', fontsize=16)
    ax2.legend(shadow=True, borderpad=1, fontsize=14, loc='best')
    plt.tight_layout()


    parent_dir = os.path.abspath(os.getcwd())
    path_figure = os.path.join(parent_dir, 'plots')
    fig.savefig(os.path.join(path_figure, cls_data.data_name + "_" + model.interp_method) , dpi=300)
    plt.close()


def plot_2D_1D(model, cls_data, plot_in_axis, plot_out_axis, color_map="viridis", vmin=0, vmax=1, marker_size=20):
    """ This function plots 2D input and 1D output data. By default, this function should work on original data only
    plot_in_axis: [axis1, axis2]
    plot_out_axis: [axis1]
    """

    TD_type = model.config["TD_type"]

    ## create mesh and data
    ### in normalized space, create prediction
    xmin, xmax = cls_data.x_data_minmax["min"][plot_in_axis[0]], cls_data.x_data_minmax["max"][plot_in_axis[0]]
    ymin, ymax = cls_data.x_data_minmax["min"][plot_in_axis[1]], cls_data.x_data_minmax["max"][plot_in_axis[1]]
    umin, umax = cls_data.u_data_minmax["min"][plot_out_axis[0]], cls_data.u_data_minmax["max"][plot_out_axis[0]]

    x_nds = jnp.linspace(xmin, xmax, 101, dtype=jnp.float64)
    y_nds = jnp.linspace(ymin, ymax, 101, dtype=jnp.float64)
    X,Y = jnp.meshgrid(x_nds, y_nds) # (101,101) each
    XY = jnp.dstack((X, Y)) # (101,101,2)
    if model.interp_method == "linear" or model.interp_method == "nonlinear":
        U_pred = model.vv_forward(model.params, XY) # (101,101,L)
    elif model.interp_method == "MLP":
        U_pred = model.vv_forward(model.params, model.activation, XY) # (101,101,L)
        

    
    U_exact = globals()["vv_fun_"+cls_data.data_name](XY) # (101,101,L)    

    fig = plt.figure(figsize=(10,5))
    gs = gridspec.GridSpec(1, 2)
    ax1 = fig.add_subplot(gs[0])
    ax2 = fig.add_subplot(gs[1])
    plt.subplots_adjust(wspace=0.4)  # Increase the width space between subplots

    surf1 = ax1.pcolormesh(X, Y, U_pred[:,:,plot_out_axis[0]], cmap=color_map, vmin=umin, vmax=umax)
    ax1.set_xlabel(f"x_{str(plot_in_axis[0]+1)}", fontsize=16)
    ax1.set_ylabel(f"x_{str(plot_in_axis[1]+1)}", fontsize=16)
    ax1.tick_params(axis='both', labelsize=12)
    cbar1 = fig.colorbar(surf1, shrink=0.8, aspect=20, pad=0.02)
    cbar1.set_label(f'u', fontsize=14)
    cbar1.ax.tick_params(labelsize=12)
    ax1.set_title('Prediction', fontsize=16)

    surf2 = ax2.pcolormesh(X, Y, U_exact[:,:,plot_out_axis[0]], cmap=color_map, vmin=umin, vmax=umax)
    ax2.set_xlabel(f"x_{str(plot_in_axis[0]+1)}", fontsize=16)
    ax2.set_ylabel(f"x_{str(plot_in_axis[1]+1)}", fontsize=16)
    ax2.tick_params(axis='both', labelsize=12)
    cbar2 = fig.colorbar(surf2, shrink=0.8, aspect=20, pad=0.02)
    cbar2.set_label(f'u', fontsize=14)
    cbar2.ax.tick_params(labelsize=12)
    ax2.set_title('Original function', fontsize=16)

    parent_dir = os.path.abspath(os.getcwd())
    path_figure = os.path.join(parent_dir, 'plots')
    fig.savefig(os.path.join(path_figure, cls_data.data_name + f"_{TD_type}_" + model.interp_method) , dpi=300)
    plt.close()


def plot_2D_classification(model, cls_data, plot_in_axis, plot_out_axis):

    TD_type = model.config["TD_type"]

    xmin, xmax = 0, 1
    ymin, ymax = 0, 1
    
    x_nds = jnp.linspace(xmin, xmax, 101, dtype=jnp.float64)
    y_nds = jnp.linspace(ymin, ymax, 101, dtype=jnp.float64)
    X,Y = jnp.meshgrid(x_nds, y_nds) # (101,101) each
    XY = jnp.dstack((X, Y)) # (101,101,2)
    if model.interp_method == "linear" or model.interp_method == "nonlinear":
        U_pred = model.vv_forward(model.params, XY) # (101,101,L)
    elif model.interp_method == "MLP":
        U_pred = model.vv_forward(model.params, model.activation, XY) # (101,101,L)
    U_pred_single = jnp.argmax(U_pred, axis=2)

    plt.figure(figsize=(6, 5))
    plt.set_cmap(plt.cm.Paired)
    plt.pcolormesh(X, Y, U_pred_single)


    ## debug
    all_inputs, all_labels = [], []

    for inputs, labels in model.test_dataloader:
        # Move data to CPU if it’s on GPU
        inputs = inputs.cpu().numpy()
        labels = labels.cpu().numpy()
        
        all_inputs.append(inputs)
        all_labels.append(labels)

    # Concatenate along the batch dimension
    x_data = np.concatenate(all_inputs, axis=0)
    u_data = np.concatenate(all_labels, axis=0)
    u_data_single = jnp.argmax(u_data, axis=1)
    
    plt.scatter(x_data[:,0], x_data[:,1], c=u_data_single, edgecolors='black')
    
    plt.xlabel(rf"$x_{plot_in_axis[0]+1}$", fontsize = 20)
    plt.ylabel(rf"$x_{plot_in_axis[1]+1}$", fontsize = 20)
    plt.xticks(fontsize=14)
    plt.yticks(fontsize=14)
    plt.xlim((0, 1))
    plt.ylim((0, 1))
    plt.tight_layout()
    
    ## Save plot
    parent_dir = os.path.abspath(os.getcwd())
    path_figure = os.path.join(parent_dir, 'plots')
    plt.savefig(os.path.join(path_figure, cls_data.data_name + f"_{TD_type}_" + model.interp_method) , dpi=300)
    plt.close()

    


    


"""
INN trainer
----------------------------------------------------------------------------------
Copyright (C) 2024  Chanwook Park
 Northwestern University, Evanston, Illinois, US, chanwookpark2024@u.northwestern.edu
"""

# import jax
# import jax.numpy as jnp
import numpy as np
import os, sys, csv
import pandas as pd
from typing import Sequence
from torch.utils.data import Dataset
from torchvision import datasets, transforms
# from torchvision.transforms import v2
import torch

def one_hot(labels, num_classes):
    # Initialize a zero array with shape (ndata, nclass)
    one_hot = np.zeros((len(labels), num_classes), dtype=int)
    
    # Set the appropriate elements to 1 using numpy indexing
    one_hot[np.arange(len(labels)), np.squeeze(labels)] = 1
    
    return one_hot

class Data_classification(Dataset):
    def __init__(self, data_name: str, config) -> None:
        if not os.path.exists('data'):
            os.makedirs('data')
        self.data_dir = 'data/'
        self.data_name = data_name
        self.nclass = config['DATA_PARAM']['nclass']
        self.var = self.nclass
        self.split_ratio = config['DATA_PARAM']['split_ratio']
        self.bool_normalize = config['DATA_PARAM']['bool_normalize']
        self.bool_image = config['DATA_PARAM']['bool_image']

        data_file = self.data_dir + data_name + '.csv'
        try:
            data = np.loadtxt(data_file, delimiter=",", dtype=np.float32, skiprows=1)
        except: 
            print(F"Data file {data_file} dose not exist. We will create the data.")
            data_generation_classification(data_name, config)
            data = np.loadtxt(data_file, delimiter=",", dtype=np.float32, skiprows=1)

        if 'mnist' in self.data_name:
            self.x_data_org = data[:, 1:]
            self.u_data_org = data[:, 0].astype(np.int32)
        else:
            self.x_data_org = data[:, config['DATA_PARAM']['input_col']]
            self.u_data_org = data[:, config['DATA_PARAM']['output_col']].astype(np.int32)
        self.u_data = one_hot(self.u_data_org, self.nclass) # u_data is the one-hot vector (ndata, nclass)
        self.dim = self.x_data_org.shape[1]
        
        
        if self.bool_normalize:
            if self.bool_image: # when image is provided with 0~255 integer pixel values
                self.x_data_minmax = {"min" : np.zeros(self.x_data_org.shape[1], dtype=np.int32), 
                                      "max" : np.ones(self.x_data_org.shape[1], dtype=np.int32) * 255}
                self.x_data = (self.x_data_org - self.x_data_minmax["min"]) / (self.x_data_minmax["max"] - self.x_data_minmax["min"])

            elif self.bool_image == False: # spiral classification
                self.x_data_minmax = {"min" : self.x_data_org.min(axis=0), "max" : self.x_data_org.max(axis=0)}
                self.x_data = (self.x_data_org - self.x_data_minmax["min"]) / (self.x_data_minmax["max"] - self.x_data_minmax["min"])
        
        else:
            if self.bool_image: # when image is provided with 0~1 floating pixel values; our MNIST dataset
                self.x_data_minmax = {"min" : np.zeros(self.x_data_org.shape[1], dtype=np.float64), 
                                      "max" : np.ones(self.x_data_org.shape[1], dtype=np.float64)}
                self.x_data = self.x_data_org
                
            elif self.bool_image == False: # 
                self.x_data_minmax = {"min" : self.x_data_org.min(axis=0), "max" : self.x_data_org.max(axis=0)}
                self.x_data = self.x_data_org
            
        print('loaded ',len(self.x_data_org),'datapoints from',data_name,'dataset')

    
    def __len__(self):
        return len(self.x_data_org)

    def __getitem__(self, idx):
        # image = np.array(self.dataframe.loc[idx, self.dataframe.columns != 'label'].values)
        # target = np.array(self.dataframe.loc[idx, 'label'])
        # image = image.reshape((28,28))
        # image = torch.tensor(image).unsqueeze(0)
        
        # image = image.reshape((28,28,1))
        # return np.array(image), np.array(target)
        return self.x_data[idx], self.u_data[idx]
    
def data_generation_classification(data_name: str, config):

    data_dir = './data'

    if data_name == 'mnist' or data_name == 'fashion_mnist':
        transform = transforms.ToTensor()

        if data_name == 'mnist':
            train_dataset = datasets.MNIST(root=data_dir, train=True, download=True, transform=transform)
            test_dataset = datasets.MNIST(root=data_dir, train=False, download=True, transform=transform)
        elif data_name == 'fashion_mnist':
            train_dataset = datasets.FashionMNIST(root=data_dir, train=True, download=True, transform=transform)
            test_dataset = datasets.FashionMNIST(root=data_dir, train=False, download=True, transform=transform)

        # Combine training and test datasets
        combined_data = [(int(label), *image.numpy().astype('float16').flatten()) for image, label in train_dataset]
        combined_data += [(int(label), *image.numpy().astype('float16').flatten()) for image, label in test_dataset]
        
        # Define column names for the header
        columns = ["label"] + [f"pixel{i}" for i in range(784)]
        
        # Convert to DataFrame and save as CSV
        combined_df = pd.DataFrame(combined_data, columns=columns)
        combined_df.to_csv(os.path.join(data_dir, f'{data_name}.csv'), index=False, header=True)

    elif data_name == 'spiral':
        halfSamples = 5000 # number of data points in each class, 5,000 -> total 10,000 data
        noise = 3
        N_SAMPLES = halfSamples * 2
        def genSpiral(deltaT, label, halfSamples, noise):
            points = np.zeros((halfSamples, 3), dtype=np.double)
            for i in range(halfSamples):
                r = i / halfSamples * 5 # radius
                # t = 1.75 * i / halfSamples * 2 * np.pi + deltaT # theta
                t = 3.43 * i / halfSamples * 2 * np.pi + deltaT # theta
                x = r * np.sin(t) + np.random.uniform(-0.1,0.1) * noise
                y = r * np.cos(t) + np.random.uniform(-0.1,0.1) * noise
                points[i] = np.array([x, y, label])
            return points

        points1 =genSpiral(0, 1, halfSamples, noise) # Positive examples
        points2 =genSpiral(np.pi, 0, halfSamples, noise) # Negative examples
        points = np.concatenate((points1, points2), axis=0)
        # shuffle
        indices = np.arange(N_SAMPLES)
        np.random.shuffle(indices)
        points = points[indices,:]
        df = pd.DataFrame(points, columns=['x1', 'x2', 'u'])
        
        # ## Plot ##
        # import matplotlib.pyplot as plt

        # plt.figure(figsize=(6, 5))
        # plt.set_cmap(plt.cm.Paired)
        # # plt.pcolormesh(xx, yy, ynew)
        # plt.scatter(points[:,0], points[:,1], c=points[:,2], edgecolors='black')

        # plt.xlabel(r'$p_1$', fontsize = 20)
        # plt.ylabel(r'$p_2$', fontsize = 20)
        # plt.xticks(fontsize=14)
        # plt.yticks(fontsize=14)
        
        # parent_dir = os.path.abspath(os.getcwd())
        # path_figure = os.path.join(parent_dir, 'plots')
        # plt.savefig(os.path.join(path_figure, 'dataset_spiral') , dpi=300)
        # plt.close()


        # Save the DataFrame to a CSV file
        df.to_csv(os.path.join(data_dir, f'{data_name}.csv'), index=False)

        


"""
INN trainer
----------------------------------------------------------------------------------
Copyright (C) 2024  Chanwook Park
 Northwestern University, Evanston, Illinois, US, chanwookpark2024@u.northwestern.edu
"""

import jax
import jax.numpy as jnp
import numpy as np
import os, sys, csv
import pandas as pd
from typing import Sequence
from torch.utils.data import Dataset
from scipy.stats import qmc


class Data_regression(Dataset):
    def __init__(self, data_name: str, config) -> None:
        if not os.path.exists('data'):
            os.makedirs('data')
        self.data_dir = 'data/'
        self.data_name = data_name
        self.data_size = config['DATA_PARAM']['data_size']
        self.input_col = config['DATA_PARAM']['input_col']
        self.output_col = config['DATA_PARAM']['output_col']
        self.dim = len(self.input_col) # size of input
        self.var = len(self.output_col) # size of output
        self.split_ratio = config['DATA_PARAM']['split_ratio']
        self.bool_normalize = config['DATA_PARAM']['bool_normalize']
        
        data_file = self.data_dir + data_name + '_' + str(self.data_size) + '.csv'
        try:
            data = np.loadtxt(data_file, delimiter=",", dtype=np.float64, skiprows=1)
        except: 
            print(F"Data file {data_file} dose not exist. We will create the data.")
            data_generation_regression(data_name, self.data_size, self.input_col)
            data = np.loadtxt(data_file, delimiter=",", dtype=np.float64, skiprows=1)
        
        self.x_data_org = data[:, self.input_col]
        self.u_data_org = data[:, self.output_col]
        
        if self.bool_normalize:    
            self.x_data_minmax = {"min" : self.x_data_org.min(axis=0), "max" : self.x_data_org.max(axis=0)}
            self.u_data_minmax = {"min" : self.u_data_org.min(axis=0), "max" : self.u_data_org.max(axis=0)}
            self.x_data = (self.x_data_org - self.x_data_minmax["min"]) / (self.x_data_minmax["max"] - self.x_data_minmax["min"])
            self.u_data = (self.u_data_org - self.u_data_minmax["min"]) / (self.u_data_minmax["max"] - self.u_data_minmax["min"])
        else:
            self.x_data_minmax = {"min" : self.x_data_org.min(axis=0), "max" : self.x_data_org.max(axis=0)}
            self.u_data_minmax = {"min" : self.u_data_org.min(axis=0), "max" : self.u_data_org.max(axis=0)}
            self.x_data = self.x_data_org
            self.u_data = self.u_data_org
            
        print('loaded ',len(self.x_data_org),'datapoints from',data_name,'dataset')
        
    def __len__(self):
        return len(self.x_data_org)

    def __getitem__(self, idx):
        return self.x_data[idx], self.u_data[idx]
    
    def denormalize(self, data, minmax):
        """ Denormalize both x_data and u_data
        data: (ndata, dim or var) either u_data or x_data
        minmax: dictionary that stores minmax values
        """
        data_org = (minmax["max"] - minmax["min"]) * data + minmax["min"]
        return data_org

def data_generation_regression(data_name: str, data_size: int, input_col: Sequence[int]):

    ## random sampling
    # x_data_org = jnp.array(np.random.rand(data_size, len(input_col)))

    ## Latin Hypercube sampling
    x_data_sampler = qmc.LatinHypercube(d=len(input_col))
    x_data_org = x_data_sampler.random(n=data_size)
        
    if data_name == "1D_1D_sine":
        u_data_org = v_fun_1D_1D_sine(x_data_org)
        cols = ['x1', 'u']

    elif data_name == "1D_1D_exp":
        u_data_org = v_fun_1D_1D_exp(x_data_org)
        cols = ['x1', 'u']

    elif data_name == "1D_2D_sine_exp":
        u_data_org = v_fun_1D_2D_sine_exp(x_data_org)
        cols = ['x1','u1','u2']

    elif data_name == "2D_1D_sine":
        u_data_org = v_fun_2D_1D_sine(x_data_org)
        cols = ['x1','x2','u']

    elif data_name == "2D_1D_exp":
        u_data_org = v_fun_2D_1D_exp(x_data_org)
        cols = ['x1','x2','u']

    elif data_name == "3D_1D_exp":
        u_data_org = v_fun_3D_1D_exp(x_data_org)
        cols = ['x1','x2','x3','u']

    elif data_name == "8D_1D_physics": # borehole function; use JAX to create data

        x_min = jnp.array([0.05,    100,  63_070,  990, 63.1, 700, 1120,  9_855], dtype=jnp.double)
        x_max = jnp.array([0.15, 50_000, 115_600, 1110,  116, 820, 1680, 12_045], dtype=jnp.double)
        x_data_org = x_data_org * (x_max-x_min) + x_min
        
        u_data_org = jax.vmap(fun_8D_1D_physics)(x_data_org)
        cols = ['x1','x2','x3','x4','x5','x6','x7','x8','u']

    elif data_name == "10D_5D_physics": # five physics function; use JAX to create data

        ## u1: Borehole function
        x_min = jnp.array([0.05,    100,  63_070,  990, 63.1, 700, 1120,  9_855], dtype=jnp.double)
        x_max = jnp.array([0.15, 50_000, 115_600, 1110,  116, 820, 1680, 12_045], dtype=jnp.double)
        x1_data_org = x_data_org[:,:8] * (x_max-x_min) + x_min
        
        ## u2: Piston simulation function
        x_min = jnp.array([30, 0.005, 0.002, 1000, 90_000, 290, 340], dtype=jnp.double)
        x_max = jnp.array([60, 0.020, 0.010, 5000,110_000, 296, 360], dtype=jnp.double)
        x2_data_org = x_data_org[:,:7] * (x_max-x_min) + x_min
        
        ## u3: OTL circuit function
        x_min = jnp.array([50, 25, 0.5, 1.2, 0.25, 50], dtype=jnp.double)
        x_max = jnp.array([150,70, 3.0, 2.5, 1.20,300], dtype=jnp.double)
        x3_data_org = x_data_org[:,:6] * (x_max-x_min) + x_min
        
        ## u4: Robot arm function
        x_min = jnp.array([     0.0,      0.0,      0.0,      0.0, 0.0, 0.0, 0.0, 0.0], dtype=jnp.double)
        x_max = jnp.array([2*jnp.pi, 2*jnp.pi, 2*jnp.pi, 2*jnp.pi, 1.0, 1.0, 1.0, 1.0], dtype=jnp.double)
        x4_data_org = x_data_org[:,:8] * (x_max-x_min) + x_min
        
        ## u5: Wing weight function
        x_min = jnp.array([150, 220, 6,-10*jnp.pi/180, 16, 0.5, 0.08, 2.5, 1700, 0.025], dtype=jnp.double)
        x_max = jnp.array([200, 300,10, 10*jnp.pi/180, 45, 1.0, 0.18, 6.0, 2500, 0.080], dtype=jnp.double)
        x5_data_org = x_data_org[:,:10] * (x_max-x_min) + x_min
        

        u_data_org = jax.vmap(fun_10D_5D_physics, in_axes=(0,0,0,0,0))(x1_data_org, x2_data_org, x3_data_org, x4_data_org, x5_data_org)
        cols=['x1', 'x2', 'x3', 'x4', 'x5', 'x6', 'x7', 'x8', 'x9', 'x10', 'u1', 'u2', 'u3', 'u4', 'u5']

    
    elif data_name == "IGAMapping2D":

        ## define IGA parameters
        
        x_min = jnp.array([0, 0], dtype=jnp.double)
        x_max = jnp.array([10, 10], dtype=jnp.double)
        
        x_data_org = x_data_org * (x_max-x_min) + x_min
        u_data_org = v_fun_IGAMapping2D(x_data_org)
        cols = ['x1','x2','u1','u2']


    data = np.concatenate((x_data_org, u_data_org), axis=1)
    df = pd.DataFrame(data, columns=cols)

    # Save the DataFrame to a CSV file
    parent_dir = os.path.abspath(os.getcwd())
    path_data = os.path.join(parent_dir, 'data')
    csv_filename = f"{data_name}_{data_size}.csv" # 1D input, 1D output, sine curve
    df.to_csv(os.path.join(path_data, csv_filename), index=False)




## define functions
def fun_1D_1D_exp(x_data_org):
    u_data_org =  jnp.exp(4*x_data_org**2 - 2*x_data_org - 1)
    return u_data_org.reshape(1,)
v_fun_1D_1D_exp = jax.vmap(fun_1D_1D_exp, in_axes = (0)) # output: (ndata, )
vv_fun_1D_1D_exp = jax.vmap(v_fun_1D_1D_exp, in_axes = (0)) # output: (ndata, ndata)


def fun_1D_1D_sine(x_data_org):
    u_data_org =  jnp.sin(2*jnp.pi*x_data_org)
    return u_data_org.reshape(1,)
v_fun_1D_1D_sine = jax.vmap(fun_1D_1D_sine, in_axes = (0)) # output: (ndata, )
vv_fun_1D_1D_sine = jax.vmap(v_fun_1D_1D_sine, in_axes = (0)) # output: (ndata, ndata)

def fun_1D_2D_sine_exp(x_data_org):
    u1 = jnp.sin(2*jnp.pi*x_data_org)
    u2 = jnp.exp(4*x_data_org**2 - 2*x_data_org - 1)
    return jnp.array([u1,u2], dtype=jnp.double).reshape(-1)
v_fun_1D_2D_sine_exp = jax.vmap(fun_1D_2D_sine_exp, in_axes = (0)) # output: (ndata, )
vv_fun_1D_2D_sine_exp = jax.vmap(v_fun_1D_2D_sine_exp, in_axes = (0)) # output: (ndata, ndata)

def fun_2D_1D_sine(x_data_org):
    u_data_org =  jnp.sin(x_data_org[0]) * jnp.cos(x_data_org[1])
    return u_data_org.reshape(1,)
v_fun_2D_1D_sine = jax.vmap(fun_2D_1D_sine, in_axes = (0)) # output: (ndata, )
vv_fun_2D_1D_sine = jax.vmap(v_fun_2D_1D_sine, in_axes = (0)) # output: (ndata, ndata)

def fun_2D_1D_exp(x_data_org):
    u_data_org =  jnp.exp(x_data_org[0] + 2*x_data_org[1])
    return u_data_org.reshape(1,)
v_fun_2D_1D_exp = jax.vmap(fun_2D_1D_exp, in_axes = (0)) # output: (ndata, )
vv_fun_2D_1D_exp = jax.vmap(v_fun_2D_1D_exp, in_axes = (0)) # output: (ndata, ndata)

def fun_3D_1D_exp(x_data_org):
    u_data_org =  (2*x_data_org[2]*jnp.sin(x_data_org[1]) - 3*x_data_org[0]) / jnp.exp(x_data_org[0] - x_data_org[1]**2)
    return u_data_org.reshape(1,)
v_fun_3D_1D_exp = jax.vmap(fun_3D_1D_exp, in_axes = (0)) # output: (ndata, )
vv_fun_3D_1D_exp = jax.vmap(v_fun_3D_1D_exp, in_axes = (0)) # output: (ndata, ndata)


def fun_8D_1D_physics(p): 
    p1, p2, p3, p4, p5, p6, p7, p8 = p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7]

    ## u1: Borehole function
    u = 2*jnp.pi* p1 * (p4-p6) * (jnp.log(p2/p3) * (1 + 2*(p7*p1) / (jnp.log(p2/p3)*p3**2*p8) + p1/p5))**(-1)
    return jnp.array([u], dtype=jnp.double)

def fun_10D_5D_physics(x1,x2,x3,x4,x5): 
    
    ## u1: Borehole function
    p=x1
    p1, p2, p3, p4, p5, p6, p7, p8 = p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7]
    u1 = 2*jnp.pi* p1 * (p4-p6) * (jnp.log(p2/p3) * (1 + 2*(p7*p1) / (jnp.log(p2/p3)*p3**2*p8) + p1/p5))**(-1)

    ## u2: Piston simulation function
    p=x2
    p1, p2, p3, p4, p5, p6, p7 = p[0], p[1], p[2], p[3], p[4], p[5], p[6]
    A = p5*p2 + 19.62*p1 - p4*p3/p2
    V = p2/(2*p4)*((A**2 + 4*p4*p5*p3*p6/p7)**0.5 - A)
    u2 = 2*jnp.pi* (p1/(p4 + p2**2*p5*p3*p6/p7/V**2))**0.5

    ## u3: OTL circuit function
    p=x3
    p1, p2, p3, p4, p5, p6 = p[0], p[1], p[2], p[3], p[4], p[5]
    u3 = ( ((12*p2/(p1+p2) + 0.74) * p6*(p5+9) + 11.35*p3)/(p6*(p5+9)+p3) 
                + (0.74*p3*p6*(p5+9))/((p6*(p5+9) + p3)*p4) )
    
    ## u4: Robot arm function
    p=x4
    p1, p2, p3, p4, p5, p6, p7, p8 = p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7]
    x, y = 0, 0
    for i in range(4):
        angle = 0
        for j in range(i+1):
            angle += p[j]
        x += p[i+4] * jnp.cos(angle)
        y += p[i+4] * jnp.sin(angle)
    u4 = (x**2 + y**2)**0.5

    ## u5: Wing weight function
    p=x5
    p1, p2, p3, p4, p5, p6, p7, p8, p9, p10 = p[0], p[1], p[2], p[3], p[4], p[5], p[6], p[7], p[8], p[9]
    u5 = ( 0.036*p1**0.758*p2**0.0035 * (p3/(jnp.cos(p4))**2)**0.6 * p5**0.006*p6**0.04 * (100*p7/jnp.cos(p4))**(-0.3)
        * (p8*p9)**0.49 + p1*p10 )
    
    return jnp.array([u1,u2,u3,u4,u5], dtype=jnp.double)


## IGA mapping

def in_range(xi, lb, ub):
    # lb: lower bound, floating number
    # ub: upper bound, floating number
    return jnp.heaviside(xi-lb,1) * jnp.heaviside(ub-xi, 0)

def NBasis(xi, L):
    xi /= L
    N1 = (1-xi)**2
    N2 = 2*xi*(1-xi)
    N3 = xi**2
    N_all = jnp.array([N1, N2, N3], dtype=jnp.float64)
    return N_all

def MBasis(eta, L):
    eta /= L
    M1 = (1-eta)**2
    M2 = 2*eta*(1-eta)
    M3 = eta**2
    M_all = jnp.array([M1, M2, M3], dtype=jnp.float64)
    return M_all

def Sum_fun(xieta, L, weights):
    xi, eta = xieta[0], xieta[1]
    N_all = NBasis(xi, L)
    M_all = MBasis(eta, L)
    NM_all = jnp.tensordot(N_all, M_all, axes=0) # (4,3)
    Sum = jnp.sum(NM_all * weights)
    return Sum

def fun_IGAMapping2D(xieta):
    # This basic mapping is nothing but the identity mapping
    # R_all * controlPts returns xi itselt.

    ## IGA parameters
    L = 10
    controlPts = np.zeros((3,3,2), dtype=np.double)
                
    controlPts[:,:,0] = np.array([[0,   0,  0],
                                    [10, 15, 20],
                                    [10, 15, 20]], dtype=np.float64)
    controlPts[:,:,1]= np.array([[10, 15, 20],
                                    [10, 15, 20],
                                    [0,   0,  0]], dtype=np.float64)
    controlPts = jnp.array(controlPts)
    weights = jnp.array([[1, 1, 1],
                        [0.5*jnp.sqrt(2), 0.5*jnp.sqrt(2), 0.5*jnp.sqrt(2)],
                        [1, 1, 1]], dtype=jnp.float64)

    xi, eta = xieta[0], xieta[1]
    N_all = NBasis(xi, L)
    M_all = MBasis(eta, L)
    NM_all = jnp.tensordot(N_all, M_all, axes=0) # (4,3)
    Sum = Sum_fun(xieta, L, weights)
    R_all = NM_all * weights / Sum # (4,3)
    xy = jnp.sum(R_all[:,:,None] * controlPts[:,:,:], axis=(0,1))
    return xy
v_fun_IGAMapping2D = jax.vmap(fun_IGAMapping2D, in_axes = (0))
vv_fun_IGAMapping2D = jax.vmap(v_fun_IGAMapping2D, in_axes = (0))



from setuptools import setup, find_packages

# Load the long_description from README.md
with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setup(
    name='pyinn',
    version='0.1.6',
    author='Chanwook Park',
    author_email='chanwookpark2024@u.northwestern.edu',
    description='Interpolating Neural Networks',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6'
)


"""
INN trainer
----------------------------------------------------------------------------------
Copyright (C) 2024  Chanwook Park
 Northwestern University, Evanston, Illinois, US, chanwookpark2024@u.northwestern.edu
"""

import jax
import jax.numpy as jnp
from jax import config
config.update("jax_enable_x64", True)
import optax
from functools import partial
import numpy as np
from torch.utils.data import DataLoader, random_split, Subset
import time
import torch
from sklearn.metrics import r2_score, classification_report
import importlib.util

from .model import * ## when using pyinn
# from model import * ## when debugging

if importlib.util.find_spec("GPUtil") is not None: # for linux & GPU
    ''' If you are funning on GPU, please install the following libraries on your anaconda environment via 
    $ conda install -c conda-forge humanize
    $ conda install -c conda-forge psutil
    $ conda install -c conda-forge gputil
    ''' 
    import humanize, psutil, GPUtil
    
    # memory report
    def mem_report(num, gpu_idx):
        ''' This function reports memory usage for both CPU and GPU'''
        print(f"-{num}-CPU RAM Free: " + humanize.naturalsize( psutil.virtual_memory().available ))
        
        GPUs = GPUtil.getGPUs()
        gpu = GPUs[gpu_idx]
        # for i, gpu in enumerate(GPUs):
        print('---GPU {:d} ... Mem Free: {:.0f}MB / {:.0f}MB | Utilization {:3.0f}%\n'.format(gpu_idx, gpu.memoryFree, gpu.memoryTotal, gpu.memoryUtil*100))

def get_linspace(xmin, xmax, nnode):
    return jnp.linspace(xmin,xmax,nnode, dtype=jnp.float64)
v_get_linspace = jax.vmap(get_linspace, in_axes=(0,0,None))

class Regression_INN:
    def __init__(self, cls_data, config):

        self.interp_method = config['interp_method']
        self.cls_data = cls_data
        self.config = config
        self.key = 3264

        self.nmode = int(config['MODEL_PARAM']['nmode'])
        self.nelem = int(config['MODEL_PARAM']['nelem'])
        self.nnode = self.nelem+1
        self.num_epochs = int(self.config['TRAIN_PARAM']['num_epochs_INN'])
        
        ## initialization of trainable parameters
        if cls_data.bool_normalize: # when the data is normalized
            self.grid_dms = jnp.tile(jnp.linspace(0,1,self.nnode, dtype=jnp.float64), (self.cls_data.dim,1)) # (dim,nnode)
        else: # when the data is not normalized
            self.grid_dms = v_get_linspace(cls_data.x_data_minmax["min"], cls_data.x_data_minmax["max"], self.nnode)

        
        self.params = jax.random.uniform(jax.random.PRNGKey(self.key), (self.nmode, self.cls_data.dim, 
                                            self.cls_data.var, self.nnode), dtype=jnp.double)       
        numParam = self.nmode*self.cls_data.dim*self.cls_data.var*self.nnode

        ## Define model
        if self.interp_method == "linear":
            model = INN_linear(self.grid_dms, self.config)
            self.forward = model.forward
            self.v_forward = model.v_forward
            self.vv_forward = model.vv_forward
        elif self.interp_method == "nonlinear":
            model = INN_nonlinear(self.grid_dms, self.config)
            self.forward = model.forward
            self.v_forward = model.v_forward
            self.vv_forward = model.vv_forward
        
        if self.interp_method == "linear" or self.interp_method == "nonlinear":
            print(f"------------INN {config['TD_type']} {self.interp_method} -------------")
            print(f"# of training parameters: {numParam}")


    @partial(jax.jit, static_argnames=['self']) # jit necessary
    def get_loss(self, params, x_data, u_data):
        ''' Compute MSE loss value at (m)th mode given upto (m-1)th mode solution, which is u_pred_old
        --- input ---
        u_p_modes: (nmode, dim, nnode, var)
        u_data: exact u from the data. (ndata_train, var)
        shape_vals_data, patch_nodes_data: defined in "get_HiDeNN_shape_fun"
        '''
        u_pred = self.v_forward(params, x_data) # (ndata_train, var)
        loss = ((u_pred- u_data)**2).mean()
        return loss, u_pred
    Grad_get_loss = jax.jit(jax.value_and_grad(get_loss, argnums=1, has_aux=True), static_argnames=['self'])


    def get_loss_r(self, grid_dms, params, x_data, u_data):
        ''' Compute MSE loss value for r-adaptivity
        --- input ---
        grid_dms: (dim, J)
        u_p_modes: (nmode, dim, nnode, var)
        u_data: exact u from the data. (ndata_train, var)
        shape_vals_data, patch_nodes_data: defined in "get_HiDeNN_shape_fun"
        '''
        # if self.interp_method == "linear":
        model = INN_linear(grid_dms, self.config)
            
        # elif self.interp_method == "nonlinear":
        #     model = INN_nonlinear(grid_dms, self.config)

        u_pred = model.v_forward(params, x_data) # (ndata_train, var)
        loss = ((u_pred- u_data)**2).mean()
        return loss, u_pred
    Grad_get_loss_r = jax.jit(jax.value_and_grad(get_loss_r, argnums=1, has_aux=True), static_argnames=['self'])
    

    @partial(jax.jit, static_argnames=['self']) 
    def update_optax(self, params, opt_state, x_data, u_data):
        ((loss, u_pred), grads) = self.Grad_get_loss(params, x_data, u_data)
        updates, opt_state = self.optimizer.update(grads, opt_state)
        params = optax.apply_updates(params, updates)
        return params, opt_state, loss, u_pred    
    
    def data_split(self):
        split_ratio = self.cls_data.split_ratio
        if self.config['DATA_PARAM']['bool_random_split'] == True and all(isinstance(item, float) for item in split_ratio):
            # random split with a split ratio
            generator = torch.Generator().manual_seed(42)
            split_data = random_split(dataset=self.cls_data,lengths=self.cls_data.split_ratio, generator=generator)
            if len(split_ratio) == 2:
                self.split_type="TT" # train and test
                train_data = split_data[0]
                test_data = split_data[1]
            elif len(split_ratio) == 3:
                self.split_type="TVT" # train, validation, test
                train_data = split_data[0]
                val_data = split_data[1]
                test_data = split_data[2]
                self.val_dataloader = DataLoader(test_data, batch_size=self.batch_size, shuffle=True)
            
        elif self.config['DATA_PARAM']['bool_random_split'] == False and all(isinstance(item, int) for item in split_ratio):
            # non-random split with a fixed number of data
            if len(split_ratio) == 2:
                self.split_type="TT" # train and test
                train_data = Subset(self.cls_data, list(range(split_ratio[0])))
                test_data = Subset(self.cls_data, list(range(split_ratio[0], split_ratio[0]+split_ratio[1])))
            elif len(split_ratio) == 3:
                self.split_type="TVT" # train, validation, test
                train_data = Subset(self.cls_data, list(range(split_ratio[0])))
                test_data = Subset(self.cls_data, list(range(split_ratio[0], split_ratio[0]+split_ratio[1])))
                val_data = Subset(self.cls_data, list(range(split_ratio[0]+split_ratio[1], split_ratio[0]+split_ratio[1]+split_ratio[2])))
                self.val_dataloader = DataLoader(test_data, batch_size=self.batch_size, shuffle=True)
        self.train_dataloader = DataLoader(train_data, batch_size=self.batch_size, shuffle=True)
        self.test_dataloader = DataLoader(test_data, batch_size=self.batch_size, shuffle=True)

    def get_error_metrics(self, u, u_pred, error, ndata):
        """ Get sum of squared error
        """
        if jnp.isnan(u_pred).any(): # if the prediction has nan value,
            print(f"[Error] INN prediction has NaN components")
            print(jnp.where(jnp.isnan(u_pred))[0])        

        if self.bool_denormalize:
            u = self.cls_data.denormalize(u, self.cls_data.u_data_minmax)
            u_pred = self.cls_data.denormalize(u_pred, self.cls_data.u_data_minmax)
        error += jnp.sum((u - u_pred)**2)
        ndata += len(u)
        return error, ndata # sum of squared error

    def inference(self, x_test):
            u_pred = self.forward(self.params, x_test[0]) # (ndata_train, var)
            u_pred = self.forward(self.params, x_test[0]) # (ndata_train, var)
            u_pred = self.forward(self.params, x_test[0]) # (ndata_train, var)
            start_time_inference = time.time()
            u_pred = self.forward(self.params, x_test[0]) # (ndata_train, var)
            print(f"\tInference time: {time.time() - start_time_inference:.4f} seconds")       

    def train(self):
        self.batch_size = int(self.config['TRAIN_PARAM']['batch_size'])
        self.learning_rate = float(self.config['TRAIN_PARAM']['learning_rate'])
        self.validation_period = int(self.config['TRAIN_PARAM']['validation_period'])
        self.bool_denormalize = self.config['TRAIN_PARAM']['bool_denormalize']
        self.error_type = self.config['TRAIN_PARAM']['error_type']
        self.patience = int(self.config['TRAIN_PARAM']['patience'])

        
        ## Split data and create dataloader
        self.data_split()
        
        ## Define optimizer
        params = self.params
        self.optimizer = optax.adam(self.learning_rate)
        opt_state = self.optimizer.init(params)
        
        ## Train
        start_time_train = time.time()
        errors_val = [] # validation error for every epoch
        for epoch in range(self.num_epochs):
            epoch_list_loss, epoch_list_acc = [], [] 
            start_time_epoch = time.time()    
            error, ndata = 0,0 # measure train error
            for batch in self.train_dataloader:
                
                x_train, u_train = jnp.array(batch[0]), jnp.array(batch[1])
                params, opt_state, loss_train, u_pred_train = self.update_optax(params, opt_state, x_train, u_train)
                error, ndata = self.get_error_metrics(u_train, u_pred_train, error, ndata)
                
            self.params = params
            print(f"Epoch {epoch+1}")
            if self.error_type == 'rmse':
                print(f"\tTrain RMSE: {jnp.sqrt(error/ndata):.4e}")
            elif self.error_type == 'mse':
                print(f"\tTrain MSE: {error/ndata:.4e}")
            elif self.error_type == 'accuracy':
                print(f"\tTrain Accuracy: {error/ndata*100:.2f}%")
            else:
                pass
            print(f"\tTrain took {time.time() - start_time_epoch:.4f} seconds")

            ## Validation
            if (epoch+1)%self.validation_period == 0:
                if self.split_type == "TT": # when there are only train & test data
                    self.val_dataloader = self.test_dataloader # deal test data as validation data
                error, ndata = 0,0
                for batch in self.val_dataloader:
                    x_val, u_val = jnp.array(batch[0]), jnp.array(batch[1])
                    _, _, loss_val, u_pred_val = self.update_optax(params, opt_state, x_val, u_val)
                    error, ndata = self.get_error_metrics(u_val, u_pred_val, error, ndata)
                
                if self.error_type == 'rmse':
                    print(f"\tVal RMSE: {jnp.sqrt(error/ndata):.4e}")
                    errors_val.append(jnp.sqrt(error/ndata))
                elif self.error_type == 'mse':
                    print(f"\tVal MSE: {error/ndata:.4e}")
                    errors_val.append(error/ndata)
                elif self.error_type == 'accuracy':
                    print(f"\tVal Accuracy: {error/ndata*100:.2f}%")
                    errors_val.append(error/ndata)
                else:
                    pass

            ## Check early stopping
            if len(errors_val) > self.patience:
                early_stopping = np.all(np.subtract(errors_val[-self.patience:], errors_val[-self.patience-1:-1])>0)
                if early_stopping: # break the epoch and finish training
                    print(f"\tEarly stopping at {epoch+1}-th epoch")
                    print(f"\tValidation losses of the latest epochs are {errors_val[-self.patience:]}")
                    break

        print(f"INN training took {time.time() - start_time_train:.4f} seconds")
        # if importlib.util.find_spec("GPUtil") is not None: # report GPU memory usage
        #     mem_report('After training', gpu_idx)

        ## Test 
        start_time_test = time.time()
        error, ndata = 0,0
        for batch in self.test_dataloader:
            x_test, u_test = jnp.array(batch[0]), jnp.array(batch[1])
            _, _, _, u_pred_test = self.update_optax(params, opt_state, x_test, u_test)
            error, ndata = self.get_error_metrics(u_test, u_pred_test, error, ndata)
        print("Test")
        if self.error_type == 'rmse':
            print(f"\tTest RMSE: {jnp.sqrt(error/ndata):.4e}")
        elif self.error_type == 'mse':
            print(f"\tTest MSE: {error/ndata:.4e}")
        elif self.error_type == 'accuracy':
            print(f"\tTest Accuracy: {error/ndata*100:.2f}%")
            
        print(f"\tTest took {time.time() - start_time_test:.4f} seconds") 

        ## Inference
        self.inference(x_test)

    def train_r(self):
        self.batch_size = int(self.config['TRAIN_PARAM']['batch_size']) * 100 # set r-batch as large as possible
        self.learning_rate = float(self.config['TRAIN_PARAM']['learning_rate'])
        self.validation_period = int(self.config['TRAIN_PARAM']['validation_period'])
        
        ## Split data and create dataloader
        self.data_split()
        
        ## Define variables
        grid_dms = self.grid_dms
        
        ## Train
        start_time_train = time.time()
        for batch in self.train_dataloader:
            # time_batch = time.time()
            x_train, u_train = jnp.array(batch[0]), jnp.array(batch[1])
            ((loss, u_pred), grads) = self.Grad_get_loss_r(grid_dms, self.params, x_train, u_train)

            ## Updated nodal coordinates - r-adaptivity
            grads = grads.at[:,[0,-1]].set(0) # fix boundary nodes
            elem_size = grid_dms[:,1:] - grid_dms[:,0:-1]  # (dim, J-1)
            elem_size_min = jnp.min(elem_size, axis=1) # (dim,) minimum element size for each dimension
            learning_rate = 0.05 * elem_size_min / jnp.max(jnp.abs(grads), axis=1)
            grid_dms -= learning_rate[:,None] * grads

        ## Save updated nodal coordinates 
        self.grid_dms = grid_dms
        # print(grid_dms)
        
        ## Re-define models
        if self.interp_method == "linear":
            model = INN_linear(self.grid_dms, self.config)
            self.forward = model.forward
            self.v_forward = model.v_forward
            self.vv_forward = model.vv_forward
        elif self.interp_method == "nonlinear":
            model = INN_nonlinear(self.grid_dms, self.config)
            self.forward = model.forward
            self.v_forward = model.v_forward
            self.vv_forward = model.vv_forward




class Regression_MLP(Regression_INN):
    def __init__(self, cls_data, config):
        super().__init__(cls_data, config) # prob being dropout probability
        
        self.forward = forward_MLP
        self.v_forward = v_forward_MLP
        self.vv_forward = vv_forward_MLP
        self.nlayers = config['MODEL_PARAM']["nlayers"]
        self.nneurons = config['MODEL_PARAM']["nneurons"]
        self.activation = config['MODEL_PARAM']["activation"]
        self.num_epochs = int(self.config['TRAIN_PARAM']['num_epochs_MLP'])

        ### initialization of trainable parameters
        layer_sizes = [cls_data.dim] + self.nlayers * [self.nneurons] + [cls_data.var]
        self.params = self.init_network_params(layer_sizes, jax.random.PRNGKey(self.key))
        weights, biases = 0,0
        for layer in self.params:
            w, b = layer[0], layer[1]
            weights += w.shape[0]*w.shape[1]
            biases += b.shape[0]
        print("------------MLP-------------")
        print(f"# of training parameters: {weights+biases}")

        
    def random_layer_params(self, m, n, key, scale=1e-2): # m input / n output neurons
      w_key, b_key = jax.random.split(key)
      return scale * jax.random.normal(w_key, (n, m)), scale * jax.random.normal(b_key, (n,))
    
    def init_network_params(self, sizes, key):
        keys = jax.random.split(key, len(sizes))
        return [self.random_layer_params(m, n, k) for m, n, k in zip(sizes[:-1], sizes[1:], keys)]
    
    @partial(jax.jit, static_argnames=['self']) # jit necessary
    def get_loss(self, params, x_data, u_data):
        ''' Compute loss value at (m)th mode given upto (m-1)th mode solution, which is u_pred_old
        --- input ---
        u_p_modes: (nmode, dim, nnode, var)
        u_data: exact u from the data. (ndata_train, var)
        shape_vals_data, patch_nodes_data: defined in "get_HiDeNN_shape_fun"
        '''
        
        u_pred = self.v_forward(params, self.activation, x_data) # (ndata_train, var)
        loss = ((u_pred- u_data)**2).mean()
        return loss, u_pred
    Grad_get_loss = jax.jit(jax.value_and_grad(get_loss, argnums=1, has_aux=True), static_argnames=['self'])

    def inference(self, x_test):
        u_pred = self.forward(self.params, self.activation, x_test[0]) # (ndata_train, var)
        u_pred = self.forward(self.params, self.activation, x_test[0]) # (ndata_train, var)
        u_pred = self.forward(self.params, self.activation, x_test[0]) # (ndata_train, var)
        start_time_inference = time.time()
        u_pred = self.forward(self.params, self.activation, x_test[0]) # (ndata_train, var)
        print(f"\tInference time: {time.time() - start_time_inference:.4f} seconds")    
    

class Classification_INN(Regression_INN):
    def __init__(self, cls_data, config):
        super().__init__(cls_data, config) # prob being dropout probability
        
        ## classification problem always normalize inputs between 0 and 1
        self.x_dms_nds = jnp.tile(jnp.linspace(0,1,self.nnode, dtype=jnp.float64), (self.cls_data.dim,1)) # (dim,nnode)
        
        ## initialization of trainable parameters
        self.params = jax.random.uniform(jax.random.PRNGKey(self.key), (self.nmode, self.cls_data.dim, 
                                                self.cls_data.var, self.nnode), dtype=jnp.double,
                                                minval=0.98, maxval=1.02) # for classification, we sould confine the params range
        # numParam = self.nmode*self.cls_data.dim*self.cls_data.var*self.nnode
        
        

    @partial(jax.jit, static_argnames=['self']) # jit necessary
    def get_loss(self, params, x_data, u_data):
        ''' Compute Cross Entropy loss value at (m)th mode given upto (m-1)th mode solution, which is u_pred_old
        --- input ---
        u_p_modes: (nmode, dim, nnode, var)
        u_data: exact u from the data. (ndata_train, var)
        shape_vals_data, patch_nodes_data: defined in "get_HiDeNN_shape_fun"
        '''
        
        u_pred = self.v_forward(params, x_data) # (ndata_train, var)
        prediction = u_pred - jax.scipy.special.logsumexp(u_pred, axis=1)[:,None] # (ndata, var = nclass)
        loss = -jnp.mean(jnp.sum(prediction * u_data, axis=1))
        return loss, u_pred
    Grad_get_loss = jax.jit(jax.value_and_grad(get_loss, argnums=1, has_aux=True), static_argnames=['self'])
    # Grad_get_loss = jax.value_and_grad(get_loss, argnums=1, has_aux=True)

    
    def get_error_metrics(self, u, u_pred, error, ndata):
        """ Get accuracy metrics. For regression, either RMSE or MSE will be returned.
            This function cannot be jitted because it uses scipy library
        --- input ---
        u: (ndata, nclass) integer vector that indicates class of the data
        u_train: (ndata, nclass) integer vector that indicates predicted class
        """
        u_single = jnp.argmax(u, axis=1)
        u_pred_single = jnp.argmax(u_pred, axis=1)
        report = classification_report(np.array(u_single), np.array(u_pred_single), output_dict=True, zero_division=1)
        error = report["accuracy"]
        ndata += len(u)
        return error*ndata, ndata


class Classification_MLP(Regression_MLP):

    def __init__(self, cls_data, config):
        super().__init__(cls_data, config) # prob being dropout probability

    @partial(jax.jit, static_argnames=['self']) # jit necessary
    def get_loss(self, params, x_data, u_data):
        ''' Compute Cross Entropy loss value at (m)th mode given upto (m-1)th mode solution, which is u_pred_old
        --- input ---
        u_p_modes: (nmode, dim, nnode, var)
        u_data: exact u from the data. (ndata_train, var)
        shape_vals_data, patch_nodes_data: defined in "get_HiDeNN_shape_fun"
        '''
        
        u_pred = self.v_forward(params, self.activation, x_data) # (ndata_train, var)
        prediction = u_pred - jax.scipy.special.logsumexp(u_pred, axis=1)[:,None] # (ndata, var = nclass)
        loss = -jnp.mean(jnp.sum(prediction * u_data, axis=1))
        return loss, u_pred
    Grad_get_loss = jax.jit(jax.value_and_grad(get_loss, argnums=1, has_aux=True), static_argnames=['self'])
    # Grad_get_loss = jax.value_and_grad(get_loss, argnums=1, has_aux=True)

    def get_error_metrics(self, u, u_pred, error, ndata):
        """ Get accuracy metrics. For regression, either RMSE or MSE will be returned.
            This function cannot be jitted because it uses scipy library
        --- input ---
        u: (ndata, nclass) integer vector that indicates class of the data
        u_train: (ndata, nclass) integer vector that indicates predicted class
        """
        u_single = jnp.argmax(u, axis=1)
        u_pred_single = jnp.argmax(u_pred, axis=1)
        report = classification_report(np.array(u_single), np.array(u_pred_single), output_dict=True, zero_division=1)
        error = report["accuracy"]
        ndata += len(u)
        return error*ndata, ndata


"""
INN trainer
----------------------------------------------------------------------------------
Copyright (C) 2024  Chanwook Park
 Northwestern University, Evanston, Illinois, US, chanwookpark2024@u.northwestern.edu
"""
# from pyinn import dataset_classification, dataset_regression, model, train, plot # with pyinn library
import dataset_classification, dataset_regression, model, train, plot # for debugging
from jax import config
config.update("jax_enable_x64", True)
import os
import yaml

run_types = ['regression', 'classification']
interp_methods = ['linear','nonlinear', 'MLP']

data_names_regression = ['1D_1D_sine', '2D_1D_sine','10D_5D_physics', ]
data_names_classification = ['spiral', 'mnist','fashion_mnist', ]


# %% User Set up
with open('./pyinn/settings.yaml','r') as file:
    settings = yaml.safe_load(file)

gpu_idx = settings['GPU']['gpu_idx']  # set which GPU to run on Athena
os.environ["CUDA_DEVICE_ORDER"] = "PCI_BUS_ID"  # GPU indexing
os.environ["CUDA_VISIBLE_DEVICES"] = str(gpu_idx)  # GPU indexing

for run_type in run_types:
    for interp_method in interp_methods:
        if run_type == 'regression':
            for data_name in data_names_regression:

                with open(f'./config/{data_name}.yaml','r') as file_dataConfig:
                    config = yaml.safe_load(file_dataConfig)
                    config['interp_method'] = interp_method
                    config['TD_type'] = settings['PROBLEM']["TD_type"]
                    config['TRAIN_PARAM']['num_epochs_INN'] = int(config['TRAIN_PARAM']['validation_period'] * 2)
                    config['TRAIN_PARAM']['num_epochs_MLP'] = int(config['TRAIN_PARAM']['validation_period'] * 2)

                # --------------------- Regression --------------------------3
                    
                ## data import
                data = dataset_regression.Data_regression(data_name, config)

                ## train
                if interp_method == "linear" or interp_method == "nonlinear":
                    regressor = train.Regression_INN(data, config)  
                elif interp_method == "MLP":
                    regressor = train.Regression_MLP(data, config)  
                regressor.train()  # Train module

                ## plot
                plot.plot_regression(regressor, data, config)

                # --------------------- Classification --------------------------
                
        if run_type == 'classification':
            for data_name in data_names_classification:
                if ('mnist' in data_name and interp_method=='nonlinear'):
                    break
                    
                with open(f'./config/{data_name}.yaml','r') as file_dataConfig:
                    config = yaml.safe_load(file_dataConfig)
                    config['interp_method'] = interp_method
                    config['TD_type'] = settings['PROBLEM']["TD_type"]
                    config['TRAIN_PARAM']['num_epochs_INN'] = int(config['TRAIN_PARAM']['validation_period'] * 2)
                    config['TRAIN_PARAM']['num_epochs_MLP'] = int(config['TRAIN_PARAM']['validation_period'] * 2)

                ## data import
                data = dataset_classification.Data_classification(data_name, config)
                
                ## train
                if interp_method == "linear" or interp_method == "nonlinear":
                    classifier = train.Classification_INN(data, config)  # HiDeNN-TD regressor class
                elif interp_method == "MLP":
                    classifier = train.Classification_MLP(data, config)  # HiDeNN-TD regressor class    
                classifier.train()  # Train module

                ## plot
                plot.plot_classification(classifier, data, config)




    






import numpy as np
import jax
import jax.numpy as jnp
from itertools import product, combinations
from functools import partial
from jax.numpy import (asarray, broadcast_arrays, can_cast,
                       empty, nan, searchsorted, where, zeros)
from jax._src.tree_util import register_pytree_node
from jax._src.numpy.util import check_arraylike, promote_dtypes_inexact
from scipy.sparse import csc_matrix


class LinearInterpolator:
  def __init__(self, grid):
    """ 1D linear interpolation
    --- input --- 
    grid: (J,) 1D vector of the grid
    """

    self.grid = grid
    # self.grid_sort = grid.at[-1].add(0.0001)

  def __call__(self, xi, values):
    
    ielem, norm_distance = self._find_ielem(xi)
    result = self._interpolate(ielem, norm_distance, values)
    return result

  def _interpolate(self, ielem, norm_distance, values):
    
    value = values[ielem] * (1-norm_distance) + values[ielem+1] * (norm_distance)
    return value

  def _find_ielem(self, xi):
    i = searchsorted(self.grid, xi) - 1
    i = where(i < 0, 0, i) # i: element index
    norm_distance = (xi - self.grid[i]) / (self.grid[i + 1] - self.grid[i])
    return i, norm_distance


class NonlinearInterpolator(LinearInterpolator):
  def __init__(self, grid,
                    nelem, nnode, s_patch, alpha_dil, p_order, 
                    mbasis, radial_basis, activation):
    super().__init__(grid) # prob being dropout probability

    self.nelem = nelem
    self.nnode = nnode
    self.s_patch = s_patch
    self.alpha_dil = alpha_dil
    self.p_order = p_order
    self.mbasis = mbasis
    self.radial_basis = radial_basis
    self.activation = activation
    
    ## compute G_inv
    self.nodes_per_elem = 2
    self.d_c = 1.0/self.nelem     # characteristic length in physical coord.
    self.a_dil = self.alpha_dil * self.d_c

    self.Elem_nodes_host = np.zeros([self.nelem, self.nodes_per_elem], dtype=np.int64)
    for j in range(1, self.nnode):
      self.Elem_nodes_host[j-1, 0] = j-1
      self.Elem_nodes_host[j-1, 1] = j 

    
    self.indices, self.indptr = self.get_adj_mat()
    self.edex_max, self.ndex_max = self.get_dex_max()
    print(f'edex_max / ndex_max: {self.edex_max} / {self.ndex_max}')
    
    # start_patch = time.time()
    (self.Elemental_patch_nodes_st, self.edexes,
      self.Nodal_patch_nodes_st, self.Nodal_patch_nodes_bool, self.Nodal_patch_nodes_idx, self.ndexes
      ) = self.get_patch_info()       
    
    self.Gs = self.vv_get_G(self.ndexes, self.Nodal_patch_nodes_st, self.Nodal_patch_nodes_bool, grid.reshape(-1,1)) # (nelem, npe=2, ndex_max+mbasis, ndex_max+mbasis)
    # self.Gs_inv = jnp.array(np.linalg.inv(self.Gs))
    # print(self.Gs)
    # print(self.Gs.shape)
    self.Gs_inv = jnp.linalg.inv(self.Gs)
    # print(self.Gs_inv)
    # print(self.Gs.shape)
    


  def __call__(self, xi, values):
    """
    --- values: (J,) nodal values
    """
  
    ielem, norm_distance = self._find_ielem(xi)
    result = self.get_phi(xi, ielem, 0, values) * (1-norm_distance) + self.get_phi(xi, ielem, 1, values) * (norm_distance)
    return result

  def get_phi(self, xi, ielem, inode, values):
    Ginv = self.Gs_inv[ielem, inode]  # (K+P+1, K+P+1) where K: ndex_max
    nodal_patch_nodes = self.Nodal_patch_nodes_st[ielem,inode]
    RP = self.Compute_RadialBasis_1D(xi.reshape(1), self.grid[nodal_patch_nodes].reshape(-1,1), 
                                     self.ndexes[ielem,inode], self.Nodal_patch_nodes_bool[ielem,inode]) # (K+P+1)
    RP_Ginv = jnp.tensordot(RP, Ginv, axes=(0,1))[:self.ndex_max]
    result = jnp.sum(RP_Ginv * jnp.take(values, nodal_patch_nodes), keepdims=False)
    return result

  def get_adj_mat(self):
    ''' Compute Adjacency matrix from the Graph theory
    --- input ---
    Elem_nodes_host: elemental connectivity stored in CPU as numpy array, (nelem, nodes_per_elem)
    nnode: number of nodes, scalar integer
    s_patch: patch size s, scalar integer
    --- output ---
    indices, indptr: indices and index pointers used in scipy spare matrix library.
    '''
    
    
    # get adjacency matrix of graph theory based on nodal connectivity
    adj_rows, adj_cols = [], []
    
    # self 
    for inode in range(self.nnode):
      adj_rows += [inode]
      adj_cols += [inode]
    
    for ielem, elem_nodes in enumerate(self.Elem_nodes_host):
      for (inode, jnode) in combinations(list(elem_nodes), 2):
        adj_rows += [inode, jnode]
        adj_cols += [jnode, inode]
    adj_values = np.ones(len(adj_rows), dtype=np.int64)
    adj_rows = np.array(adj_rows, dtype=np.int64)
    adj_cols = np.array(adj_cols, dtype=np.int64)
    
    # build sparse matrix
    adj_sp = csc_matrix((adj_values, 
                              (adj_rows, adj_cols)),
                            shape = (self.nnode, self.nnode))
    adj_s = csc_matrix((adj_values, 
                              (adj_rows, adj_cols)),
                            shape = (self.nnode, self.nnode))
    
    # compute s th power of the adjacency matrix to get s th order of connectivity
    for itr in range(self.s_patch-1):
      adj_s = adj_s.dot(adj_sp)
    indices = adj_s.indices
    indptr = adj_s.indptr
    
    return indices, indptr


  def get_dex_max(self): # delete d_c, XY, nnode
    ''' For padding algorithm in GPU computation, we need to know the maximum number of elemental patch nodes and 
    nodal patch nodes, edex_max and ndex_max, respectively.
    --- input ---
    indices, indptr: sparce matrix form of the Adjacency matrix
    s_patch: patch size s
    Elem_nodes: elemental connectivity stored in CPU as numpy array, (nelem, nodes_per_elem)
    nelem, nodes_per_elem, dim: scalar integers..
    --- output ---
    edex_max: maximum number of elemental patch nodes
    ndex_max: maximum number of nodal patch nodes
    '''
    
    dim = 1
    edex_max = (2+2*self.s_patch)**dim # estimated value of edex_max
    edexes = np.zeros(self.nelem, dtype=np.int64) # (num_elements, )
    ndexes = np.zeros((self.nelem, self.nodes_per_elem), dtype=np.int64) # static, (nelem, nodes_per_elem)
    
    for ielem, elem_nodes in enumerate(self.Elem_nodes_host):
      if len(elem_nodes) == 2 and dim == 1: # 1D Linear element
        nodal_patch_nodes0 = self.indices[ self.indptr[elem_nodes[0]] : self.indptr[elem_nodes[0]+1] ] # global index, # node_idx 0
        nodal_patch_nodes1 = self.indices[ self.indptr[elem_nodes[1]] : self.indptr[elem_nodes[1]+1] ] # global index
        ndexes[ielem, :] = np.array([len(nodal_patch_nodes0),len(nodal_patch_nodes1)])
        elemental_patch_nodes = np.unique(np.concatenate((nodal_patch_nodes0, nodal_patch_nodes1)))  

      edexes[ielem] = len(elemental_patch_nodes)
    edex_max = np.max(edexes)
    ndex_max = np.max(ndexes)
    return edex_max, ndex_max

  def get_patch_info(self): # for block, delete s_patch, d_c, XY
    ''' Inside the C-HiDeNN shape function builder, compute the patch information for each node on each element.
    --- input ---
    indices, indptr: sparce matrix form of the Adjacency matrix
    edex_max: maximum number of elemental patch nodes
    ndex_max: maximum number of nodal patch nodes
    Elem_nodes: elemental connectivity stored in CPU as numpy array, (nelem, nodes_per_elem)
    nelem, nodes_per_elem, dim: scalar integers..
    --- output ---
    Elemental_patch_nodes_st: elemental patch nodes, (nelem, edex_max)
    edexes: number of patch nodes for each element, (nelem,)
    Nodal_patch_nodes_st: nodal patch nodes for each element, (nelem, nodes_per_elem, ndex_max)
    Nodal_patch_nodes_bool: the same shape as Nodal_patch_nodes_st, returns 1 if the node has non-zero convolution patch function value, 
                            returns 0 if the node has zero convolution patch function value.
    Nodal_patch_nodes_idx: the same shape as Nodal_patch_nodes_st, stores nodal index in the corresponding elemental patch nodes
    ndexes: number of patch nodes for each node, and each element, (nelem, nodes_per_elem)
    '''    
    dim = 1
    
    # Assign memory to variables
    ## Elemental patch
    Elemental_patch_nodes_st = np.zeros((self.nelem, self.edex_max), dtype=np.int64) # edex_max should be grater than 100!
    edexes = np.zeros(self.nelem, dtype=np.int64) # (num_elements, )
    ## Nodal patch
    Nodal_patch_nodes_st = (-1)*np.ones((self.nelem, self.nodes_per_elem, self.ndex_max), dtype=np.int64) # static, (nelem, nodes_per_elem, ndex_max)
    Nodal_patch_nodes_bool = np.zeros((self.nelem, self.nodes_per_elem, self.ndex_max), dtype=np.int64) # static, (nelem, nodes_per_elem, ndex_max)
    Nodal_patch_nodes_idx = (-1)*np.ones((self.nelem, self.nodes_per_elem, self.ndex_max), dtype=np.int64) # static, (nelem, nodes_per_elem, ndex_max)
    ndexes = np.zeros((self.nelem, self.nodes_per_elem), dtype=np.int64) # static, (nelem, nodes_per_elem)
    
    
    for ielem, elem_nodes in enumerate(self.Elem_nodes_host):
        
      # 1. for loop: nodal_patch_nodes in global nodal index
      for inode_idx, inode in enumerate(elem_nodes):
        nodal_patch_nodes = np.sort(self.indices[ self.indptr[elem_nodes[inode_idx]] : self.indptr[elem_nodes[inode_idx]+1] ]) # global index
        ndex = len(nodal_patch_nodes)
        ndexes[ielem, inode_idx] = ndex
        Nodal_patch_nodes_st[ielem, inode_idx, :ndex] = nodal_patch_nodes  # global nodal index
        Nodal_patch_nodes_bool[ielem, inode_idx, :ndex] = np.where(nodal_patch_nodes>=0, 1, 0)
      
      
      # 2. get elemental_patch_nodes    
      if len(elem_nodes) == 2 and dim == 1: # 1D Linear element
        elemental_patch_nodes = np.unique(np.concatenate((Nodal_patch_nodes_st[ielem, 0, :ndexes[ielem, 0]],
                                                            Nodal_patch_nodes_st[ielem, 1, :ndexes[ielem, 1]])))  # node_idx 1
          
      
      edex = len(elemental_patch_nodes)
      edexes[ielem] = edex
      Elemental_patch_nodes_st[ielem, :edex] = elemental_patch_nodes
      
      # 3. for loop: get nodal_patch_nodes_idx
      for inode_idx, inode in enumerate(elem_nodes):
        nodal_patch_nodes_idx = np.searchsorted(
            elemental_patch_nodes, Nodal_patch_nodes_st[ielem, inode_idx, :ndexes[ielem, inode_idx]]) # local index
        Nodal_patch_nodes_idx[ielem, inode_idx, :ndexes[ielem, inode_idx]] = nodal_patch_nodes_idx
            
    # Convert everything to device array
    Elemental_patch_nodes_st = jnp.array(Elemental_patch_nodes_st)
    edexes = jnp.array(edexes)
    Nodal_patch_nodes_st = jnp.array(Nodal_patch_nodes_st)
    Nodal_patch_nodes_bool = jnp.array(Nodal_patch_nodes_bool)
    Nodal_patch_nodes_idx = jnp.array(Nodal_patch_nodes_idx)
    ndexes = jnp.array(ndexes)
    
    return Elemental_patch_nodes_st, edexes, Nodal_patch_nodes_st, Nodal_patch_nodes_bool, Nodal_patch_nodes_idx, ndexes


  def in_range(self, xi, lb, ub):
    ''' Returns 1 when lb < xi < ub, 0 otherwise '''
    # lb: lower bound, floating number
    # ub: upper bound, floating number
    return jnp.heaviside(xi-lb,1) * jnp.heaviside(ub-xi, 0)

  @partial(jax.jit, static_argnames=['self'])
  def get_R_cubicSpline(self, xy, xvi, a_dil):
    ''' Cubic spline radial basis function '''
    zI = jnp.linalg.norm(xy - xvi)/a_dil
    R = ((2/3 - 4*zI**2 + 4*zI**3         ) * self.in_range(zI, 0.0, 0.5) +    # phi_i
                        (4/3 - 4*zI + 4*zI**2 - 4/3*zI**3) * self.in_range(zI, 0.5, 1.0))
    return R
  v_get_R_cubicSpline = jax.vmap(get_R_cubicSpline, in_axes = (None, None,0,None))

  @partial(jax.jit, static_argnames=['self'])
  def get_R_gaussian1(self, xy, xvi, a_dil):
    ''' Gaussian radial basis function, zI^1 '''
    zI = jnp.linalg.norm(xy - xvi)/a_dil
    R = jnp.exp(-zI)
    return R
  v_get_R_gaussian1 = jax.vmap(get_R_gaussian1, in_axes = (None, None,0,None))

  @partial(jax.jit, static_argnames=['self'])
  def get_R_gaussian2(self, xy, xvi, a_dil):
    ''' Gaussian radial basis function, zI^2 '''
    zI = jnp.linalg.norm(xy - xvi)/a_dil
    R = jnp.exp(-zI**2)
    return R
  v_get_R_gaussian2 = jax.vmap(get_R_gaussian2, in_axes = (None, None,0,None))

  @partial(jax.jit, static_argnames=['self'])
  def get_R_gaussian3(self, xy, xvi, a_dil):
    ''' Gaussian radial basis function, zI^3 '''
    zI = jnp.linalg.norm(xy - xvi)/a_dil
    R = jnp.exp(-zI**3)
    return R
  v_get_R_gaussian3 = jax.vmap(get_R_gaussian3, in_axes = (None, None,0,None))

  @partial(jax.jit, static_argnames=['self'])
  def get_R_gaussian4(self, xy, xvi, a_dil):
    ''' Gaussian radial basis function, zI^4 '''
    zI = jnp.linalg.norm(xy - xvi)/a_dil
    R = jnp.exp(-zI**4)
    return R
  v_get_R_gaussian4 = jax.vmap(get_R_gaussian4, in_axes = (None, None,0,None))

  @partial(jax.jit, static_argnames=['self'])
  def get_R_gaussian5(self, xy, xvi, a_dil):
    ''' Gaussian radial basis function, zI^5 '''
    zI = jnp.linalg.norm(xy - xvi)/a_dil
    R = jnp.exp(-zI**5)
    return R
  v_get_R_gaussian5 = jax.vmap(get_R_gaussian5, in_axes = (None, None,0,None))
  
  @partial(jax.jit, static_argnames=['self'])
  def get_R_cosine(self, xy, xvi, a_dil):
    ''' Cubic spline radial basis function '''
    zI = jnp.linalg.norm(xy - xvi)/a_dil
    R = 0.5* ( jnp.cos(zI) + 1 ) * self.in_range(zI, 0.0, jnp.pi)
    return R
  v_get_R_cosine = jax.vmap(get_R_cosine, in_axes = (None, None,0,None))
  
  

  @partial(jax.jit, static_argnames=['self']) # This will slower the function
  def Compute_RadialBasis_1D(self, xy, xv, ndex, nodal_patch_nodes_bool):
    """ 
    --- input ---
    # xy: point of interest (dim,)
    # xv: ndoal coordinates of patch nodes. ()
    # ndex: number of nodse in the nodal patch
    # ndex_max: max of ndex, precomputed value
    # nodal_patch_nodes_bool: boolean vector that tells ~~~
    # a_dil: dilation parameter for cubic spline
    # mbasis: number of polynomial terms
    --- output ---
    RP: [R(x), P(x)] vector, R(x) is the radial basis, P(x) is the polynomial basis, (ndex_max + m_basis,)
    """
    
    RP = jnp.zeros(self.ndex_max + self.mbasis, dtype=jnp.double)
    
    if self.radial_basis == 'cubicSpline':
        RP = RP.at[:self.ndex_max].set(self.v_get_R_cubicSpline(xy, xv, self.a_dil) * nodal_patch_nodes_bool)
    if self.radial_basis == 'gaussian1':
        RP = RP.at[:self.ndex_max].set(self.v_get_R_gaussian1(xy, xv, self.a_dil) * nodal_patch_nodes_bool)        
    if self.radial_basis == 'gaussian2':
        RP = RP.at[:self.ndex_max].set(self.v_get_R_gaussian2(xy, xv, self.a_dil) * nodal_patch_nodes_bool)        
    if self.radial_basis == 'gaussian3':
        RP = RP.at[:self.ndex_max].set(self.v_get_R_gaussian3(xy, xv, self.a_dil) * nodal_patch_nodes_bool)        
    if self.radial_basis == 'gaussian5':
        RP = RP.at[:self.ndex_max].set(self.v_get_R_gaussian3(xy, xv, self.a_dil) * nodal_patch_nodes_bool)
    if self.radial_basis == 'cosine':
        RP = RP.at[:self.ndex_max].set(self.v_get_R_cosine(xy, xv, self.a_dil) * nodal_patch_nodes_bool)
    
    
    if self.activation == 'polynomial':
      if self.mbasis > 0: # 1st
        RP = RP.at[self.ndex_max   : self.ndex_max+ 2].set(jnp.array([1 , xy[0] ]))   # N 1, x
          
      if self.mbasis > 2: # 2nd
        RP = RP.at[self.ndex_max+ 2: self.ndex_max+ 3].set(jnp.array([xy[0]**2]))   # N x^2
          
      if self.mbasis > 3: # 3rd
        RP = RP.at[self.ndex_max+ 3: self.ndex_max+ 4].set(jnp.array([xy[0]**3]))   # N x^3
          
      if self.mbasis > 4: # 4th
        RP = RP.at[self.ndex_max+ 4: self.ndex_max+ 5].set(jnp.array([xy[0]**4]))   # N x^4
      
      if self.mbasis > 5: # 4th
        RP = RP.at[self.ndex_max+ 5: self.ndex_max+ 6].set(jnp.array([xy[0]**5]))   # N x^5
      
    elif self.activation == 'sinusoidal':
      if self.mbasis > 0: # 1st
        RP = RP.at[self.ndex_max   : self.ndex_max+ 2].set(jnp.array([1 , jnp.sin(xy[0]) ]))   # N 1, sin(x)
          
      if self.mbasis > 2: # 2nd
        P = RP.at[self.ndex_max+ 2: self.ndex_max+ 3].set(jnp.sin(2*xy[0]))   # N sin(2x)
          
      if self.mbasis > 3: # 3rd
        RP = RP.at[self.ndex_max+ 3: self.ndex_max+ 4].set(jnp.sin(3*xy[0]))   # N sin(3x)
    
    elif self.activation == 'exponential':
      if self.mbasis > 0: # 1st
        RP = RP.at[self.ndex_max   : self.ndex_max+ 2].set(jnp.array([1 , jnp.exp(xy[0]) ]))   # N 1, exp(x)
          
      if self.mbasis > 2: # 2nd
        RP = RP.at[self.ndex_max+ 2: self.ndex_max+ 3].set(jnp.exp(2*xy[0]))   # N exp(2x)
          
      if self.mbasis > 3: # 3rd
        RP = RP.at[self.ndex_max+ 3: self.ndex_max+ 4].set(jnp.exp(3*xy[0]))   # N exp(3x)
    
    elif self.activation == 'sigmoid':
      if self.mbasis > 0: # 1st
        RP = RP.at[self.ndex_max   : self.ndex_max+ 2].set(jnp.array([1 , jax.nn.sigmoid(xy[0]) ]))   # N 1, exp(x)
          
      if self.mbasis > 2: # 2nd
        RP = RP.at[self.ndex_max+ 2: self.ndex_max+ 3].set(jax.nn.sigmoid(2*xy[0]))   # N exp(2x)
          
      if self.mbasis > 3: # 3rd
        RP = RP.at[self.ndex_max+ 3: self.ndex_max+ 4].set(jax.nn.sigmoid(3*xy[0]))   # N exp(3x)
    
    elif self.activation == 'tanh':
      if self.mbasis > 0: # 1st
        RP = RP.at[self.ndex_max   : self.ndex_max+ 2].set(jnp.array([1 , jax.nn.tanh(xy[0]) ]))   # N 1, exp(x)
          
      if self.mbasis > 2: # 2nd
        RP = RP.at[self.ndex_max+ 2: self.ndex_max+ 3].set(jax.nn.tanh(2*xy[0]))   # N exp(2x)
          
      if self.mbasis > 3: # 3rd
        RP = RP.at[self.ndex_max+ 3: self.ndex_max+ 4].set(jax.nn.tanh(3*xy[0]))   # N exp(3x)
    
    elif self.activation == 'gelu':
      if self.mbasis > 0: # 1st
        RP = RP.at[self.ndex_max   : self.ndex_max+ 2].set(jnp.array([1 , jax.nn.gelu(xy[0]) ]))   # N 1, exp(x)
          
      if self.mbasis > 2: # 2nd
        RP = RP.at[self.ndex_max+ 2: self.ndex_max+ 3].set(jax.nn.gelu(2*xy[0]))   # N exp(2x)
          
      if self.mbasis > 3: # 3rd
        RP = RP.at[self.ndex_max+ 3: self.ndex_max+ 4].set(jax.nn.gelu(3*xy[0]))   # N exp(3x)
    return RP

  v_Compute_RadialBasis_1D = jax.vmap(Compute_RadialBasis_1D, in_axes = (None, 0,None,None,None), out_axes=1)
  v2_Compute_RadialBasis_1D = jax.vmap(Compute_RadialBasis_1D, in_axes = (None, None,0,0,0), out_axes=0)

  @partial(jax.jit, static_argnames=['self']) # unneccessary
  def get_G(self, ndex, nodal_patch_nodes, nodal_patch_nodes_bool, XY):
    ''' Compute assembled moment matrix G. Refer to Section 2.2 of:
        "Park, Chanwook, et al. "Convolution hierarchical deep-learning neural network 
        (c-hidenn) with graphics processing unit (gpu) acceleration." Computational Mechanics (2023): 1-27."
    --- input ---
    ndex: number of nodal patch nodes, scalar integer
    nodal_patch_nodes: nodal patch nodes, (ndex_max,)
    nodal_patch_nodes_bool: nodal patch nodes boolean, (ndex_max,)
    XY: nodal coorinates, (nnode, dim)
    ndex_max: maximum number of ndexes
    a_dil: dilation parameter a, scalar double
    mbasis: number of polynomial basis functions, scalar integer
    radial_basis: type of radial basis function, string
    --- output ---
    G: assembled moment matrix, (ndex_max + m_basis, ndex_max + m_basis)
    '''
    # nodal_patch_nodes_bool: (ndex_max,)
    G = jnp.zeros((self.ndex_max + self.mbasis, self.ndex_max + self.mbasis), dtype=jnp.double)
    xv = XY[nodal_patch_nodes,:]
    RPs = self.v_Compute_RadialBasis_1D(xv, xv, ndex, nodal_patch_nodes_bool) # (ndex_max + mbasis, ndex_max)
    
    G = G.at[:,:self.ndex_max].set(RPs * nodal_patch_nodes_bool[None,:])                        
    
    # Make symmetric matrix
    G = jnp.tril(G) + jnp.triu(G.T, 1)
    
    # Build diagonal terms to nullify dimensions
    Imat = jnp.eye(self.ndex_max) * jnp.abs(nodal_patch_nodes_bool-1)[:,None]
    G = G.at[:self.ndex_max,:self.ndex_max].add(Imat)
    return G # G matrix
  v_get_G = jax.vmap(get_G, in_axes = (None, 0,0,0,None))
  vv_get_G = jax.vmap(v_get_G, in_axes = (None, 0,0,0,None))

     

     
